# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Manage configuration for pvman via the CLI."""

import typer

from pvman.cli.commands.config.extend import config_extend_app
from pvman.cli.commands.config.get import config_get_app
from pvman.cli.commands.config.set import config_set_app
from pvman.cli.commands.config.unset import config_unset_app

config_app = typer.Typer(
    no_args_is_help=True,
    rich_markup_mode="rich",
    help=":gear: Perform operations with pvman configuration.",
)

config_app.add_typer(config_get_app)
config_app.add_typer(config_set_app)
config_app.add_typer(config_unset_app)
config_app.add_typer(config_extend_app)



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Store or override values in the pvman configuration."""

from typing import Annotated

from pathlib import Path

from nebulog import logger

import typer

from pvman.cli.helpers import BasicConverter as Text
from pvman.config import ConfigurationDomain, resolve_app_manager
from pvman.logging import setup_app_logging

config_set_app = typer.Typer(no_args_is_help=True)


HELP_MSG = (
    ":floppy_disk: Store a key in the configuration file.\n\n"
    ":rotating_light: [bold red]NOTE:[/] To store a [b][i]negative[/i][/b] number in "
    "the configuration, you must pass the double dash separator to prevent the "
    "application from interpreting the value as a flag:\n\n"
    "[bold yellow]$[/] [green]pvman[/] config set somekey "
    "[blue]--[/] -1"
)


@config_set_app.command(name="set", help=HELP_MSG)
def set_command(
    key: Annotated[
        str, typer.Argument(help=":key: The configuration key to be stored.")
    ],
    value: Annotated[
        Text,
        typer.Argument(
            help=":keycap_#: The value to be stored with the key.", parser=Text
        ),
    ],
    path: Annotated[
        Path, typer.Option(help=":bus_stop: Specify a custom configuration file.")
    ] = None,
    secret: Annotated[
        bool,
        typer.Option(
            "--secret",
            "-s",
            help=":lock: Store configuration in the secret manager instead.",
        ),
    ] = False,
) -> None:
    """Store a key in the configuration file."""
    setup_app_logging(debug=False)
    domain = ConfigurationDomain.from_flag(is_secret=secret)

    logger.info(
        "Storing configuration key via CLI",
        key=key,
        value=value.output,
        is_secret=secret,
    )

    if path is not None:  # pragma: no cover
        logger.info("Using custom configuration file", config=path)

    app_manager = resolve_app_manager(domain, path)

    if value.output is None:
        typer.echo(f'Could not parse the value "{value.input}"', err=True)

        raise typer.Exit(1)

    app_manager[domain.value, key] = value.output
    app_manager.save(domain.value)

    logger.debug("pvman exited successfully")



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Convert string inputs for CLI.

Typer doesn't support arguments or options with more than one type, requiring users to
address the limitation with their own solutions. Having been generated with Galactipy,
pvman chooses to implement a simple API to do basic type conversions.

There are alternatives to this implementation:

1. A GitHub PR (https://github.com/fastapi/typer/pull/723) is currently open to
   integrate common Pydantic types natively into Typer, but has not been yet merged to
   the main branch;
2. `pydantic-typer` (https://github.com/pypae/pydantic-typer) is a plugin that enables
   Pydantic support for Typer.

"""

from ast import literal_eval

from nebulog import logger

PrimitiveTypes = str | int | float | bool | list | tuple | dict
NonBooleanPrimitives = str | int | float | list | tuple | dict


class BasicConverter:
    """Convert string inputs to appropriate Python types.

    Parameters
    ----------
    value : str
        A string passed to a Typer CLI command for parsing.
    """

    BOOLEAN_VALUES = ("true", "false")
    NUMERIC_PREFIXES = "+-"
    CONTAINER_PREFIXES = "[{("
    CONTAINER_TYPES = (list, tuple, dict)

    def __init__(self, value: str) -> None:
        """Initialise the converter, storing input value and converted output value."""
        self.input = value
        self.output = self._convert_value(self.input)

    def _convert_value(self, value: str) -> PrimitiveTypes | None:
        """Parse a string into a boolean, numeric or container type."""
        if not value:  # Empty string
            return None

        if value.casefold() in self.BOOLEAN_VALUES:
            return literal_eval(value.capitalize())

        return self._parse_by_first_character(value)

    def _parse_by_first_character(self, value: str) -> NonBooleanPrimitives | None:
        """Parse a string by its first character, either a numeric or container type."""
        first_char = value[0]

        if first_char.isdigit() or first_char in self.NUMERIC_PREFIXES:
            return self.parse_numeric_value(value)

        if first_char in self.CONTAINER_PREFIXES:
            return self.parse_container_value(value)

        # Return as string if no other type matches
        return value

    @staticmethod
    def parse_numeric_value(value: str) -> str | int | float | None:
        """Parse a string into a numeric value, whether integer, float or scientific."""
        # Try integer first (more specific check)
        if "." not in value and "e" not in value.lower():
            try:
                return int(value)

            except ValueError:
                pass

        # Try float
        try:
            return float(value)

        except ValueError:
            return value

    @staticmethod
    def parse_container_value(value: str) -> list | tuple | dict | None:
        """Parse a string into a valid container value.

        Will parse strings into lists, tuples and dictionaries as-is, but will convert
        sets into tuples, as sets are not valid for TOML configuration files.
        """  # noqa: DOC201
        try:
            result = literal_eval(value)

        except (ValueError, SyntaxError):
            return None

        else:
            if isinstance(result, set):
                logger.warning(
                    f"Parsed a {type(result).__name__} object, will convert to tuple"
                )
                return tuple(result)

            return result

    def __str__(self) -> str:
        """Return the string representation of the parsed value for printing."""
        return str(self.output)

    def __repr__(self) -> str:
        """Return the string representation of the parsed value."""
        return repr(self.output)



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Specify Rich styles to render text output in the terminal.

Themes defined in this module can be added to the `AppCustomStyles` class and used in
conjunction with `rich.console.Console` to override the default terminal theme.

See https://rich.readthedocs.io/en/stable/style.html#style-themes for more information.

"""

from dataclasses import dataclass

from rich.theme import Theme

# Noctis VS Code theme text colours, full colorscheme definition at:
# https://github.com/liviuschera/noctis
NOCTIS_THEME = Theme(
    {
        "string": "#49e9a6",
        "fstring": "#16b673",
        "comment": "#5b858b",
        "function": "#16a6b6",
        "method": "#49d6e9",
        "standout": "#49ace9",
        "number": "#7060eb",
        "keyword": "#df769b",
        "declaration": "#e66533",
        "property": "#d67e5c",
        "constant": "#d5971a",
        "variable": "#e4b781",
        "bold string": "bold #49e9a6",
        "bold fstring": "bold #16b673",
        "bold comment": "bold #5b858b",
        "bold function": "bold #16a6b6",
        "bold method": "bold #49d6e9",
        "bold standout": "bold #49ace9",
        "bold number": "bold #7060eb",
        "bold keyword": "bold #df769b",
        "bold declaration": "bold #e66533",
        "bold property": "bold #d67e5c",
        "bold constant": "bold #d5971a",
        "bold variable": "bold #e4b781",
    }
)


@dataclass(frozen=True)
class AppCustomThemes:
    """Define custom Rich themes for console output with pvman."""

    NOCTIS: Theme = NOCTIS_THEME



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Custom logger formatting utilities for pvman."""

from pvman.logging.text_tools import LoggerFormatter

LOGGER_FORMATTER = LoggerFormatter()

_FILE_BASE_TEMPLATE = (
    "{time:%Y-%m-%d %H:%M:%S}.{time:SSS} | "
    "{elapsed} | "
    "{level:<8} | "  # Based on the default levels in Loguru (max. 8 in "CRITICAL")
    "{location_string} | "
    "{message:<60} ¶"
)

FILE_FORMATTER_WITH_EXTRA = _FILE_BASE_TEMPLATE + " {extra}\n{exception}"
FILE_FORMATTER_WITHOUT_EXTRA = _FILE_BASE_TEMPLATE + "\n{exception}"


def _format_location_string(record: dict) -> str:
    """Extract and format location information as `file:line`."""
    file_name = record["file"].name
    line_number = record["line"]
    padding = LOGGER_FORMATTER.get_location_padding()

    return f"{file_name}:{line_number}".ljust(padding)


def file_formatter(record: dict) -> str:
    """Define the default formatting for Loguru calls to be stored in the log files.

    Parameters
    ----------
    record : dict
        The Loguru record dictionary with the information about the logging context.

    Returns
    -------
    str
        A formatted string with standardised padding for the individual parts of the log
        entry.
    """
    record.update({"location_string": _format_location_string(record)})

    if record["extra"]:
        return FILE_FORMATTER_WITH_EXTRA

    return FILE_FORMATTER_WITHOUT_EXTRA



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Parse custom loggers defined for pvman."""

import re
from datetime import datetime

FILE_PARSER = re.compile(
    r"""
        (?P<ts>\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}.\d{3})
        \s\|\s
        \d:\d{2}:\d{2}.\d{6}
        \s\|\s
        (?P<level>\w+)
        \s+\|\s
        (?P<module>\w+)\.py:(?P<line>\d+)
        \s+\|\s
        (?P<message>\S.*?)
        \s*¶  # Extra is deliberately omitted, use JSON serialised logger instead
    """,
    re.VERBOSE,
)


def log_caster(groups: dict) -> None:  # pragma: no cover
    """Convert the parts of a parsed log record line into convenient Python types.

    Parameters
    ----------
    groups : dict
        The Loguru record dictionary with the information about the logging context.
    """
    groups["ts"] = datetime.strptime(groups["ts"], "%Y-%m-%d %H:%M:%S.%f")
    groups["line"] = int(groups["line"])



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Configuration manager for pvman."""

from typing import Literal, Self

from pathlib import Path

from orbittings import Nucleus

from pvman.config.constants import generate_default_config_schema, get_default_config
from pvman.config.mappings import ConfigurationDomain, EnvvarPrefix


class AppManager(Nucleus):
    """Extend the Orbittings `Nucleus` class to manage pvman configuration.

    Creates an instance of a `Nucleus` object with preset default contents and base
    directory, both defined in the `constants` module.
    """

    def __init__(self):
        """Initialise the manager by calling the `Nucleus` superclass."""
        super().__init__(
            default_contents=generate_default_config_schema(),
            base_dir=get_default_config(),
        )

    @classmethod
    def default(cls) -> Self:
        """Provide an `AppManager` instance with default settings and secrets files."""
        instance = cls()
        instance.add("settings", envvar_prefix=EnvvarPrefix.SETTINGS.value)
        instance.add("secrets", envvar_prefix=EnvvarPrefix.SECRETS.value)

        return instance

    @classmethod
    def custom(
        cls, path: Path, domain: Literal["settings", "secrets"] | ConfigurationDomain
    ) -> Self:
        """Provide a manager instance with a single domain and custom definitions.

        Parameters
        ----------
        path : Path
            Path to a valid custom TOML configuration file for pvman.
        domain : {"settings", "secrets"}
            Flag to treat the given `path` as a settings or secrets file.

        Returns
        -------
        AppManager
            A manager instance with a single domain and custom definitions.
        """
        instance = cls()
        domain = ConfigurationDomain(domain)

        if domain == ConfigurationDomain.SETTINGS:
            instance.add(
                domain.value,
                custom_config=path,
                envvar_prefix=EnvvarPrefix.SETTINGS.value,
            )

        else:
            instance.add(
                domain.value,
                custom_config=path,
                envvar_prefix=EnvvarPrefix.SECRETS.value,
            )

        return instance



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Launch the pvman interface as a module script."""

from pvman.config import AppManager
from pvman.tui.main_window import TerminalApp

APP_MANAGER = AppManager.default()

app = TerminalApp(APP_MANAGER.settings.theme)
app.run()



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Helper functions for streamlining pvman functionality."""

from typing import TYPE_CHECKING, Literal

from pathlib import Path

from pvman.config.manager import AppManager

if TYPE_CHECKING:
    from orbittings import Nucleus

    from pvman.config.mappings import ConfigurationDomain


def resolve_app_manager(
    domain: "Literal['settings', 'secrets'] | ConfigurationDomain",
    custom_path: Path | None = None,
) -> "Nucleus":
    """Resolve the Orbittings Manager to be used for other operations.

    Parameters
    ----------
    domain : {"settings", "secrets"}
        Flag to treat the given `path` as a settings or secrets file.
    custom_path : Path, optional
        A custom path to load a configuration from.

    Returns
    -------
    Nucleus
        A manager object for accessing global settings or secrets.
    """
    if custom_path is None:
        return AppManager.default()

    return AppManager.custom(custom_path, domain)



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Retrieve foundational values for enabling standard behaviour for pvman."""

from pathlib import Path

from platformdirs import user_config_path, user_log_path

from pvman import __version__


def get_default_config() -> Path:
    """Retrieve the default configuration path for pvman.

    Returns
    -------
    Path
        A Path object pointing to the User Config Path where configuration for
        pvman will be stored.
    """
    return user_config_path("pvman")


def get_default_log_path(filename: str | Path) -> Path:
    """Retrieve the default path to store pvman logs.

    Parameters
    ----------
    filename : str, Path
        Name of the file to be created inside the User Log Path.

    Returns
    -------
    Path
        A Path object pointing to the User Log Path where a file for pvman logs
        will be stored.
    """
    return user_log_path("pvman") / filename


def generate_default_config_schema():
    """Create the default configuration schema for pvman.

    Returns
    -------
    dict
        A dictionary containing the keys and values for the vanilla pvman
        configuration, wrapped inside a function to avoid mutation issues.
    """
    return {
        "VERSION": __version__,
        "THEME": "noctis",
        # UPDATEME with future default sections to be included
    }



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Helper functions designed to print unconventional types in Rich consoles."""

from rich.pretty import pprint

from dynaconf.utils.boxing import DynaBox
from dynaconf.vendor.box.box_list import BoxList


def pretty_print_setting(manager, key, config_type):
    """Print a Orbittings setting as a literal value on a Rich console.

    Parameters
    ----------
    manager : Nucleus
        A Orbittings manager instance.
    key : str, None
        The key to retrieve from `config_type`.
    config_type : str
        The OrbitalSystem object to inspect for `key`.
    """
    if key is None:
        _print_entire_config(manager, config_type)
    else:
        _print_specific_setting(manager, key, config_type)


def _print_entire_config(manager, config_type):
    """Print the entire Orbittings configuration as a dictionary on a Rich console."""
    config_dict = manager[config_type].to_dict()

    pprint(config_dict)


def _print_specific_setting(manager, key, config_type):
    """Print a Orbittings setting with their adequate type on a Rich console."""
    value = manager[config_type, key]
    formatted_value = _format_value_for_printing(value)

    pprint(formatted_value)


def _format_value_for_printing(value):
    """Format box and list Dynaconf objects to their respective literals."""
    if isinstance(value, DynaBox):
        return value.to_dict()

    if isinstance(value, BoxList):
        return value.to_list()

    return value



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


# UPDATEME With additional components in `tui/components/`
# See Textual documentation at:
# https://textual.textualize.io/tutorial/
"""Assemble a Textual application for a terminal user interface."""

from typing import ClassVar

from pathlib import Path

from nebulog import logger

from textual.app import App, ComposeResult
from textual.widgets import Footer, Label

from art import text2art

from pvman.tui.themes import AppCustomThemes

CSS_DIRECTORY = Path(__file__).parent / "css"


class TerminalApp(App):
    """Textual app to serve as the pvman interface.

    Parameters
    ----------
    theme : str
        Theme name to launch the application with.
    """

    CSS_PATH: ClassVar = [
        CSS_DIRECTORY / "demo.tcss",  # UPDATEME by removing when no longer needed
        CSS_DIRECTORY / "noctis.tcss",
    ]

    def __init__(self, theme: str) -> None:
        """Initialise the Terminal User Interface."""
        super().__init__()

        logger.info("Launching pvman TUI", theme=theme)
        self.default_theme = theme

    def compose(self) -> ComposeResult:  # noqa: PLR6301
        """Create child widgets for the app."""
        # UPDATEME by replacing with your own widgets
        yield Label(text2art("pvman", "tarty1"), classes="title")
        yield Label(
            "[i][b]PVM analysis like you never seen before.[/]", classes="description"
        )
        yield Footer()

    def on_mount(self) -> None:
        """Execute instructions when launching the interface."""
        logger.info("Mounting the application interface")

        for theme in AppCustomThemes:
            self.register_theme(theme.value)

        self.theme = self.default_theme


if __name__ == "__main__":
    app = TerminalApp("noctis")
    app.run()



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Launch the pvman interface."""

from typing import Annotated

from pathlib import Path

from nebulog import logger

import typer

from pvman.config import resolve_app_manager
from pvman.logging import setup_app_logging
from pvman.tui.main_window import TerminalApp

launch_app = typer.Typer(rich_markup_mode="rich")


@launch_app.command()
def launch(
    config: Annotated[
        Path,
        typer.Option(
            "--config",
            "-c",
            help=(
                ":bus_stop: Specify a custom configuration file to launch the "
                "application."
            ),
        ),
    ] = None,
    debug: Annotated[
        bool,
        typer.Option(
            "--debug",
            "-d",
            help=(
                ":bug: Log operations to the terminal at the "
                "[b][logging.level.debug]DEBUG[/logging.level.debug][/b] level."
            ),
        ),
    ] = False,
) -> None:
    """:pager: Launch the pvman interface."""
    setup_app_logging(debug=debug)

    logger.info("Calling pvman via CLI")

    if config is not None:  # pragma: no cover
        logger.info("Using custom configuration file", config=config)

    app_manager = resolve_app_manager("settings", config)

    interface = TerminalApp(app_manager.settings.theme)

    interface.run()

    logger.debug("pvman exited successfully")



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Remove values from the pvman configuration."""

from typing import Annotated

from pathlib import Path

from nebulog import logger

import typer

from pvman.config import ConfigurationDomain, resolve_app_manager
from pvman.logging import setup_app_logging

config_unset_app = typer.Typer(no_args_is_help=True)


@config_unset_app.command()
def unset(
    key: Annotated[
        str, typer.Argument(help=":key: The configuration key to be removed.")
    ],
    path: Annotated[
        Path, typer.Option(help=":bus_stop: Specify a custom configuration file.")
    ] = None,
    secret: Annotated[
        bool,
        typer.Option(
            "--secret",
            "-s",
            help=":lock: Retrieve configuration from the secret manager instead.",
        ),
    ] = False,
) -> None:
    """:fire: [red]Remove[/] a top-level key from the configuration."""
    setup_app_logging(debug=False)
    domain = ConfigurationDomain.from_flag(is_secret=secret)

    logger.info("Removing configuration key via CLI", key=key, is_secret=secret)

    if path is not None:  # pragma: no cover
        logger.info("Using custom configuration file", config=path)

    app_manager = resolve_app_manager(domain, path)

    app_manager.unset(f"{domain.value}.{key}")
    app_manager.save(domain.value)

    logger.debug("pvman exited successfully")



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Define logging capabilities for pvman."""

from nebulog import install, logger

from pvman.config.constants import get_default_log_path
from pvman.logging.formatters import file_formatter


def setup_app_logging(*, debug: bool = False) -> None:
    """Configure the available loggers at pvman runtime.

    Parameters
    ----------
    debug : bool, default False
        Define whether logs will be called at the DEBUG level or a higher level
        depending on the type of logger.
    """
    file_log_level = "DEBUG" if debug else "INFO"
    shell_log_level = "DEBUG" if debug else "WARNING"

    install(level=shell_log_level)

    logger.enable("")

    logger.add(
        get_default_log_path("app.log"),
        format=file_formatter,
        level=file_log_level,
        rotation="10 MB",
        retention="30 days",
        backtrace=False,
    )

    # Special log file for bug reporting, containing only the last run produced
    logger.add(
        get_default_log_path("report.log"),
        format=file_formatter,
        level="TRACE",
        mode="w",
        backtrace=False,
    )

    logger.add(
        get_default_log_path("app.jsonl"),
        serialize=True,
        rotation="10 MB",
        retention="30 days",
        backtrace=False,
    )


__all__ = ["file_formatter", "setup_app_logging"]



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


# UPDATEME With subcommand apps in `cli/commands/`, see documentation at:
# https://typer.tiangolo.com/tutorial/
# See recommended configuration for multicommand applications at:
# https://typer.tiangolo.com/tutorial/one-file-per-command/#main-module-mainpy
"""PVM analysis like you never seen before."""

from typing import Annotated

import typer
from rich.console import Console

from pvman import __version__
from pvman.cli.commands.config import config_app
from pvman.cli.commands.launch import launch_app
from pvman.cli.styling import AppCustomThemes

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
app.add_typer(launch_app)
app.add_typer(config_app, name="config")


def version_callback(print_version: bool) -> None:
    """Print the program version in a Rich console with the Noctis theme."""
    if print_version:
        Console(theme=AppCustomThemes.NOCTIS).print(
            f":package:[declaration]pvman[/] [bold fstring]{__version__}[/]"
        )

        raise typer.Exit


@app.callback()
def main(
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            "-v",
            help=":bulb: Print the current version of this program and exit.",
            callback=version_callback,
            is_eager=True,
        ),
    ] = False,
) -> None:
    """PVM analysis like you never seen before.

    See below for commands and options.
    """
    pass



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Specify Textual themes to be loadable and rendered in the TUI application.

Themes defined in this module can be added to the `AppCustomThemes` class and registered
to the TUI app, see https://textual.textualize.io/guide/design/ for more information.

"""

from typing import TypedDict

from enum import Enum

from textual.theme import Theme


class NoctisAuxColours(TypedDict):
    """Typed dictionary for shared Noctis colours across themes."""

    secondary: str
    warning: str
    error: str
    success: str
    accent: str


# Noctis VS Code theme colours, full colorscheme definition at:
# https://github.com/liviuschera/noctis
NOCTIS_AUX_COLOURS_DARK: NoctisAuxColours = {
    "secondary": "#e4b781",
    "warning": "#d5971a",
    "error": "#e66533",
    "success": "#16b673",
    "accent": "#7060eb",
}

NOCTIS_AUX_COLOURS_LIGHT: NoctisAuxColours = {
    "secondary": "#d67e5c",
    "warning": "#e4b781",
    "error": "#e66533",
    "success": "#16b673",
    "accent": "#16a3b6",
}

NOCTIS_THEME = Theme(
    name="noctis",
    primary="#169fb1",
    foreground="#a5b5b5",
    background="#052529",
    surface="#083d44",
    panel="#041d20",
    boost="#083d44ee",
    dark=True,
    variables={"border": "#062e32"},
    **NOCTIS_AUX_COLOURS_DARK,
)

NOCTIS_AZUREUS_THEME = Theme(
    name="noctis-azureus",
    primary="#1679b6",
    foreground="#9fb6c6",
    background="#07273b",
    surface="#09334e",
    panel="#062132",
    boost="#003c61ee",
    dark=True,
    variables={"border": "#0a3652"},
    **NOCTIS_AUX_COLOURS_DARK,
)

NOCTIS_MINIMUS_THEME = Theme(
    name="noctis-minimus",
    primary="#496d83",
    foreground="#96a8b6",
    background="#1b2932",
    surface="#202e37",
    panel="#17232b",
    boost="#1d3544ee",
    dark=True,
    variables={"border": "#243742"},
    **NOCTIS_AUX_COLOURS_DARK,
)

NOCTIS_BORDO_THEME = Theme(
    name="noctis-bordo",
    primary="#997582",
    foreground="#bbaab0",
    background="#322a2d",
    surface="#413036",
    panel="#2c2528",
    boost="#453138aa",
    dark=True,
    variables={"border": "#8f566a33"},
    **NOCTIS_AUX_COLOURS_DARK,
)

NOCTIS_UVA_THEME = Theme(
    name="noctis-uva",
    primary="#6e67a8",
    foreground="#a9a5c0",
    background="#292640",
    surface="#2f2c49",
    panel="#232136",
    boost="#35305aee",
    dark=True,
    variables={"border": "#363253"},
    **NOCTIS_AUX_COLOURS_DARK,
)

NOCTIS_VIOLA_THEME = Theme(
    name="noctis-viola",
    primary="#8767a8",
    foreground="#b3a5c0",
    background="#30243d",
    surface="#3d2e4d",
    panel="#2b2136",
    boost="#402d53ee",
    dark=True,
    variables={"border": "#422b5a"},
    **NOCTIS_AUX_COLOURS_DARK,
)

NOCTIS_LUX_THEME = Theme(
    name="noctis-lux",
    primary="#0099ad",
    foreground="#888477",
    background="#fef8ec",
    surface="#f6edda",
    panel="#f0e9d6",
    boost="#d1ebefcc",
    dark=False,
    variables={"border": "#d1e8eb"},
    **NOCTIS_AUX_COLOURS_LIGHT,
)

NOCTIS_LILAC_THEME = Theme(
    name="noctis-lilac",
    primary="#7060eb",
    foreground="#75718e",
    background="#f2f1f8",
    surface="#e9e7f3",
    panel="#e2dff6",
    boost="#d5d2ef99",
    dark=False,
    variables={"border": "#c9c2f9"},
    **NOCTIS_AUX_COLOURS_LIGHT,
)

NOCTIS_HIBERNUS_THEME = Theme(
    name="noctis-hibernus",
    primary="#0099ad",
    foreground="#71838e",
    background="#f4f6f6",
    surface="#e1eeef",
    panel="#caedf2",
    boost="#d1ebefcc",
    dark=True,
    variables={"border": "#d1e8eb"},
    **NOCTIS_AUX_COLOURS_LIGHT,
)


class AppCustomThemes(Enum):
    """Define custom Textual themes for rendering pvman with."""

    NOCTIS = NOCTIS_THEME

    AZUREUS = NOCTIS_AZUREUS_THEME
    MINIMUS = NOCTIS_MINIMUS_THEME
    BORDO = NOCTIS_BORDO_THEME
    UVA = NOCTIS_UVA_THEME
    VIOLA = NOCTIS_VIOLA_THEME

    LUX = NOCTIS_LUX_THEME
    LILAC = NOCTIS_LILAC_THEME
    HIBERNUS = NOCTIS_HIBERNUS_THEME



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Perform basic string operations to refine logger formatting."""

from functools import lru_cache
from pathlib import Path

from nebulog import logger

logger.disable("pvman")


class LoggerFormatter:
    """Handle formatting calculations for logger output alignment.

    Parameters
    ----------
    project_root : Path, optional
        Root directory to scan for Python files. Defaults to parent directory of this
        file.
    """

    ADDITIONAL_CHARACTERS = 4  # Additional characters for ".py" and ":" separator
    SHELL_LJUST_TEMPLATE = "[YYYY.mm.dd HH:MM:SS]+sss x critical x "
    SHELL_RJUST_TEMPLATE = " x "  # Separator between message and location
    MINIMUM_WRAPPER_WIDTH = 20

    def __init__(self, project_root: Path | None = None) -> None:
        """Initialize formatter with project root directory."""
        self.project_root = (
            project_root if project_root is not None else Path(__file__).parents[1]
        )

    @property
    @lru_cache(maxsize=1)
    def python_modules(self) -> tuple[Path, ...]:
        """Get all Python modules in the project.

        Returns
        -------
        tuple
            A tuple of Path objects containing all modules of `project_root`.
        """
        return tuple(self.project_root.glob("**/*.py"))

    @lru_cache(maxsize=1)
    def get_longest_filename(self) -> int:
        """Find the longest filename in the project.

        Returns
        -------
        int
            The length of the longest filename inside `project_root`.
        """
        if not self.python_modules:  # pragma: no cover
            return 0

        return max(len(module.stem) for module in self.python_modules)

    @staticmethod
    def _get_file_linecount(filepath: Path) -> int:
        """Count the number of lines in a single file."""
        try:
            with filepath.open("r", encoding="utf-8") as file:
                return sum(1 for _ in file)
        except (OSError, UnicodeDecodeError) as e:  # pragma: no cover
            logger.warning(f"Could not read file {filepath}: {e}")
            return 0

    @lru_cache(maxsize=1)
    def get_project_linecount_oom(self) -> int:
        """Find the largest order of magnitude of line counts in the project.

        Returns
        -------
        int
            Number of digits of the largest line count.
        """
        if not self.python_modules:  # pragma: no cover
            return 1

        largest_linecount = max(
            self._get_file_linecount(file) for file in self.python_modules
        )

        return len(str(largest_linecount))

    def get_location_padding(self) -> int:
        """Calculate the width necessary to turn location into an aligned column.

        Location is a combination of logging information comprised of `filename:line`.

        Returns
        -------
        int
            Total padding width needed for location alignment.
        """
        return (
            self.get_longest_filename()
            + self.get_project_linecount_oom()
            + self.ADDITIONAL_CHARACTERS
        )



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Mapping objects for the pvman configuration manager."""

from enum import Enum


class EnvvarPrefix(Enum):
    """Environment variable prefixes for pvman configuration."""

    SETTINGS = "PVMAN"
    SECRETS = "PVMAN_SECRET"


class ConfigurationDomain(Enum):
    """Configuration domain identifiers."""

    SETTINGS = "settings"
    SECRETS = "secrets"

    @classmethod
    def from_flag(cls, *, is_secret: bool) -> "ConfigurationDomain":
        """Return the appropriate domain based on the secret flag.

        Parameters
        ----------
        is_secret : bool
            Whether the configuration is secret/sensitive.

        Returns
        -------
        ConfigurationDomain
            The corresponding configuration domain.
        """
        return cls.SECRETS if is_secret else cls.SETTINGS



# pvman
# Copyright (C) 2025 The Nummertopia Delegation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Append values to existing arrays in the pvman configuration."""

from typing import Annotated

from pathlib import Path

from nebulog import logger

import typer

from pvman.cli.helpers import BasicConverter as Text
from pvman.config import ConfigurationDomain, resolve_app_manager
from pvman.logging import setup_app_logging

config_extend_app = typer.Typer(no_args_is_help=True)


HELP_MSG = (
    ":straight_ruler: Extend an array key in the configuration file.\n\n"
    "If no setting exists for the key, you cancreate an array with the single value "
    "provided.\n\n"
    ":rotating_light: [bold red]NOTE:[/] To extend a value in the configuration with a "
    "[b][i]negative[/i][/b] number, you must pass the double dash separator to prevent "
    "the application from interpreting it as a flag:\n\n"
    "[bold yellow]$[/] [green]pvman[/] config extend somekey "
    "[blue]--[/] -1"
)


@config_extend_app.command(name="extend", help=HELP_MSG)
def extend_command(
    key: Annotated[
        str, typer.Argument(help=":key: The configuration key to be extended.")
    ],
    value: Annotated[
        Text,
        typer.Argument(
            help=":keycap_#: The value to be stored with the key.", parser=Text
        ),
    ],
    path: Annotated[
        Path, typer.Option(help=":bus_stop: Specify a custom configuration file.")
    ] = None,
    secret: Annotated[
        bool,
        typer.Option(
            "--secret",
            "-s",
            help=":lock: Store configuration in the secret manager instead.",
        ),
    ] = False,
    create: Annotated[
        bool,
        typer.Option(
            "--create-on-missing",
            "-c",
            help=(
                ":new: Add the provided value in an array if the setting is not set. "
                "Will raise an error otherwise."
            ),
        ),
    ] = False,
) -> None:
    """Extend an array key in the configuration file."""
    setup_app_logging(debug=False)
    domain = ConfigurationDomain.from_flag(is_secret=secret)

    logger.info(
        "Extending array key via CLI", key=key, value=value.output, is_secret=secret
    )

    if path is not None:  # pragma: no cover
        logger.info("Using custom configuration file", config=path)

    app_manager = resolve_app_manager(domain, path)

    if value.output is None:
        typer.echo(f'Could not parse the value "{value.input}"', err=True)

        raise typer.Exit(1)

    current_setting = app_manager.get(f"{domain.value}.{key}")

    if current_setting is None and create:
        app_manager[domain.value, key] = [value.output]

        app_manager.save(domain.value)
        exit_code = 0

    elif current_setting is None:
        typer.echo(
            f"Setting `{key}` was not found, if you wish to update the configuration, "
            "run the command with the `--create-on-missing` option",
            err=True,
        )
        exit_code = 1

    elif isinstance(current_setting, list):
        current_setting.append(value.output)
        app_manager[domain.value, key] = current_setting

        app_manager.save(domain.value)
        exit_code = 0

    else:
        typer.echo(
            f"To extend settings, the value for `{key}` must be an array, got "
            f"{current_setting} instead",
            err=True,
        )
        exit_code = 1

    if exit_code != 0:
        logger.debug("pvman exited successfully")

    else:
        logger.debug("pvman exited with an error")

    raise typer.Exit(exit_code)
